from __future__ import annotations

import asyncio
import json
import logging
import shutil
import tempfile
from abc import ABC
from asyncio import Task, CancelledError
from collections import defaultdict
from contextlib import suppress
from datetime import timedelta, datetime
from multiprocessing import Process, Queue
from pathlib import Path
from queue import Empty
from typing import Optional, Union, Any, Generator, List, AsyncIterator, Dict

import aiofiles
from aiostream import stream
from aiostream.core import Stream
from attrs import define

from resotocore.analytics import AnalyticsEventSender, InMemoryEventSender, AnalyticsEvent
from resotocore.async_extensions import run_async
from resotocore.core_config import CoreConfig
from resotocore.db.db_access import DbAccess
from resotocore.db.deferred_edge_db import PendingDeferredEdges
from resotocore.db.graphdb import GraphDB
from resotocore.db.model import GraphUpdate
from resotocore.error import ImportAborted
from resotocore.ids import TaskId, GraphName
from resotocore.message_bus import MessageBus, CoreMessage
from resotocore.model.graph_access import GraphBuilder
from resotocore.model.model_handler import ModelHandlerDB, ModelHandler
from resotocore.service import Service
from resotocore.system_start import db_access, setup_process, reset_process_start_method
from resotocore.types import Json
from resotocore.util import utc, uuid_str, shutdown_process

log = logging.getLogger(__name__)


class ProcessAction(ABC):
    """
    Base class to exchange commands between processes.
    Important: all messages must be serializable in order to get pickled/unpickled.
    """


@define
class ReadFile(ProcessAction):
    """
    Read a file from disk.
    """

    path: Path
    task_id: Optional[str]

    def jsons(self) -> Generator[Json, Any, None]:
        with open(self.path, "r", encoding="utf-8") as f:
            for line in f:
                if line.strip():
                    yield json.loads(line)


@define
class ReadElement(ProcessAction):
    """
    Read an incoming element:
    - either a line of text or
    - a complete json element
    Parent -> Child: for every incoming data line.
    """

    elements: List[Union[bytes, Json]]
    task_id: Optional[str]

    def jsons(self) -> Generator[Json, Any, None]:
        return (e if isinstance(e, dict) else json.loads(e) for e in self.elements)


@define
class MergeGraph(ProcessAction):
    """
    Merge the graph that has been read so far.
    Parent -> Child: once EOF of the incoming graph is reached.
    """

    graph: GraphName
    change_id: str
    is_batch: bool = False
    task_id: Optional[TaskId] = None


@define
class EmitAnalyticsEvent(ProcessAction):
    """
    Emit this message to the event sender.
    Child -> Parent: to have the event from the child process propagated to the parent process.
    """

    event: AnalyticsEvent


class PoisonPill(ProcessAction):
    """
    Sentence the process to die.
    Parent -> Child: when the update process is interrupted.
    """


@define
class Result(ProcessAction):
    result: Union[GraphUpdate, str]

    def get_value(self) -> GraphUpdate:
        if isinstance(self.result, str):
            raise ImportAborted(self.result)
        else:
            return self.result


BatchSize = 100000


class DbUpdaterProcess(Process):
    """
    This update class implements Process and is supposed to run as separate process.
    Note: default starting method is supposed to be "spawn".

    This process has 2 queues to read input from and write output to.
    All elements in either queues are of type ProcessAction.

    The parent process should stream the raw commands of graph to this process via ReadElement objects.
    Once the MergeGraph action is received, the graph gets imported.
    From here the parent expects result messages from the child.
    All events happen in the child are forwarded to the parent via EmitEvent.
    Once the graph update is done, a result is send.
    The result is either an exception in case of failure or a graph update in success case.
    """

    def __init__(
        self,
        read_queue: Queue[ProcessAction],
        write_queue: Queue[ProcessAction],
        config: CoreConfig,
        graph_name: GraphName,
        change_id: str,
    ) -> None:
        super().__init__(name="merge_update")
        self.read_queue = read_queue
        self.write_queue = write_queue
        self.config = config
        self.change_id = change_id
        self.graph_name = graph_name

    def next_action(self) -> ProcessAction:
        try:
            # graph is read into memory. If the sender does not send data in a given amount of time,
            # we raise an exception and abort the update.
            return self.read_queue.get(True, 90)
        except Empty as ex:
            raise ImportAborted("Merge process did not receive any data for more than 90 seconds. Abort.") from ex

    async def merge_graph(self, db: DbAccess) -> GraphUpdate:  # type: ignore
        model_handler = ModelHandlerDB(db)
        model = await model_handler.load_model(self.graph_name)
        builder = GraphBuilder(model, self.change_id)
        nxt = self.next_action()
        if isinstance(nxt, ReadFile):
            for element in nxt.jsons():
                builder.add_from_json(element)
            nxt = self.next_action()
        elif isinstance(nxt, ReadElement):
            while isinstance(nxt, ReadElement):
                for element in nxt.jsons():
                    builder.add_from_json(element)
                log.debug(f"Read {int(BatchSize / 1000)}K elements in process")
                nxt = self.next_action()
        if isinstance(nxt, PoisonPill):
            log.debug("Got poison pill - going to die.")
            shutdown_process(0)
        elif isinstance(nxt, MergeGraph):
            log.debug("Graph read into memory")
            builder.check_complete()
            graphdb = db.get_graph_db(nxt.graph)
            outer_edge_db = db.pending_deferred_edge_db
            await graphdb.insert_usage_data(builder.usage)
            _, result = await graphdb.merge_graph(builder.graph, model, nxt.change_id, nxt.is_batch)
            # sizes of model entries have been adjusted during the merge. Update the model in the db.
            await model_handler.update_model(graphdb.name, list(model.kinds.values()))
            if nxt.task_id and builder.deferred_edges:
                await outer_edge_db.update(PendingDeferredEdges(nxt.task_id, utc(), nxt.graph, builder.deferred_edges))
                log.debug(f"Updated {len(builder.deferred_edges)} pending outer edges for collect task {nxt.task_id}")
            return result

    async def setup_and_merge(self) -> GraphUpdate:
        sender = InMemoryEventSender()
        _, _, sdb = DbAccess.connect(self.config.args, timedelta(seconds=3), verify=self.config.run.verify)
        db = db_access(self.config, sdb, sender)
        result = await self.merge_graph(db)
        for event in sender.events:
            self.write_queue.put(EmitAnalyticsEvent(event))
        return result

    def run(self) -> None:
        try:
            # Entrypoint of the new service
            setup_process(self.config.args, self.config)
            log.info(f"Import process started: {self.pid}")
            result = asyncio.run(self.setup_and_merge())
            self.write_queue.put(Result(result))
            log.info(f"Update process done: {self.pid}. {result} Exit.")
            shutdown_process(0)
        except Exception as ex:
            # not all exceptions can be pickled. Use string representation.
            self.write_queue.put(Result(repr(ex)))
            log.error(f"Update process interrupted. Preemptive Exit. {ex}", exc_info=ex)
            shutdown_process(1)


@define
class GraphUpdateTask:
    """
    This class represents a graph update task that is queued for processing.
    The update is written to a temporary file.
    """

    db: GraphDB
    path: Path
    deadline: datetime
    maybe_batch: Optional[str]
    task_id: Optional[TaskId]

    def cleanup(self) -> None:
        shutil.rmtree(self.path.parent)


class GraphMerger(Service):
    def __init__(
        self,
        model_handler: ModelHandler,
        event_sender: AnalyticsEventSender,
        config: CoreConfig,
        message_bus: MessageBus,
    ) -> None:
        self.model_handler = model_handler
        self.event_sender = event_sender
        self.config = config
        self.message_bus = message_bus
        self.run_lock = asyncio.Lock()
        self.running_imports: Dict[TaskId, int] = defaultdict(int)
        self.update_queue: asyncio.Queue[GraphUpdateTask] = asyncio.Queue()
        self.concurrent_updates = asyncio.Semaphore(self.config.graph_update.parallel_imports)
        self.handler_task: Optional[Task[Any]] = None

    async def __process_item(self, item: GraphUpdateTask) -> Union[GraphUpdate, Exception]:
        log.info(f"Start processing graph merge from queue: {item}")
        try:
            return await self.__merge_graph_process(item.db, item.path, item.deadline, item.maybe_batch, item.task_id)
        except Exception as ex:
            log.error(f"Failed to process graph merge: {item}", exc_info=ex)
            return ex
        finally:
            item.cleanup()

    async def start(self) -> None:
        async def wait_for_update() -> None:
            log.info("Start waiting for graph updates")
            cl = stream.cycle(stream.call(self.update_queue.get))
            fl = stream.map(cl, self.__process_item, task_limit=self.config.graph_update.parallel_imports)
            with suppress(CancelledError):
                async with fl.stream() as streamer:
                    async for update in streamer:
                        if isinstance(update, GraphUpdate):
                            log.info(f"Finished spawned graph merge: {update}")

        self.handler_task = asyncio.create_task(wait_for_update())

    async def stop(self) -> None:
        if self.handler_task:
            self.handler_task.cancel()
            with suppress(Exception):
                await self.handler_task
        # cleanup all queued entries
        while not self.update_queue.empty():
            self.update_queue.get_nowait().cleanup()

    async def merge_graph(
        self,
        db: GraphDB,
        content: AsyncIterator[Union[bytes, Json]],
        max_wait: timedelta,
        maybe_batch: Optional[str],
        task_id: Optional[TaskId],
        wait_for_result: bool = False,
    ) -> Optional[GraphUpdate]:
        # increment count
        if task_id:
            async with self.run_lock:
                self.running_imports[task_id] += 1

        deadline = utc() + max_wait
        if wait_for_result:
            return await self.__merge_graph_process(db, content, deadline, maybe_batch, task_id)
        else:
            td = Path(tempfile.mkdtemp()) / "graph"
            log.debug(f"Do not merge directly. Write to temp file: {td}")
            async with aiofiles.open(td, "wb") as f:
                async for line in content:
                    await f.write(line if isinstance(line, bytes) else (json.dumps(line) + "\n").encode("utf-8"))
            await self.update_queue.put(GraphUpdateTask(db, td, deadline, maybe_batch, task_id))
            log.debug("GraphMerge operation queued.")
            return None

    async def __merge_graph_process(
        self,
        db: GraphDB,
        content: Union[AsyncIterator[Union[bytes, Json]], Path],
        deadline: datetime,
        maybe_batch: Optional[str],
        task_id: Optional[TaskId],
    ) -> GraphUpdate:
        async with self.concurrent_updates:
            change_id = maybe_batch if maybe_batch else uuid_str()
            write: Queue[ProcessAction] = Queue()
            read: Queue[ProcessAction] = Queue()
            updater = DbUpdaterProcess(write, read, self.config, db.name, change_id)  # the process communication queue
            stale = timedelta(seconds=5).total_seconds()  # consider dead communication after this amount of time
            dead_adjusted = False

            async def send_to_child(pa: ProcessAction) -> bool:
                alive = updater.is_alive()
                if alive:
                    await run_async(write.put, pa, True, stale)
                return alive

            def read_results() -> Task[GraphUpdate]:
                async def read_forever() -> GraphUpdate:
                    nonlocal deadline
                    nonlocal dead_adjusted
                    while utc() < deadline:
                        # After exit of updater: adjust the deadline once
                        if not updater.is_alive() and not dead_adjusted:
                            log.debug("Import process done or dead. Adjust deadline.")
                            deadline = utc() + timedelta(seconds=30)
                            dead_adjusted = True
                        try:
                            action = await run_async(read.get, True, stale)
                            if isinstance(action, EmitAnalyticsEvent):
                                await self.event_sender.capture([action.event])
                            elif isinstance(action, Result):
                                return action.get_value()
                        except Empty:
                            # empty is fine
                            pass
                    raise ImportAborted(f"Import process died or deadline exceeded. (ExitCode: {updater.exitcode})")

                return asyncio.create_task(read_forever())

            task: Optional[Task[GraphUpdate]] = None
            result: Optional[GraphUpdate] = None
            try:
                reset_process_start_method()  # other libraries might have tampered the value in the mean time
                updater.start()
                task = read_results()  # concurrently read result queue
                # Either send a file or stream the content directly
                if isinstance(content, Path):
                    await send_to_child(ReadFile(content, task_id))
                else:
                    chunked: Stream = stream.chunks(content, BatchSize)
                    async with chunked.stream() as streamer:
                        async for lines in streamer:
                            if not await send_to_child(ReadElement(lines, task_id)):
                                # in case the child is dead, we should stop
                                break
                await send_to_child(MergeGraph(db.name, change_id, maybe_batch is not None, task_id))
                result = await task  # wait for final result
                await self.model_handler.load_model(db.name, force=True)  # reload model to get the latest changes
                return result
            finally:
                # update running imports and send event if completed
                if task_id:
                    async with self.run_lock:
                        self.running_imports[task_id] -= 1
                        if self.running_imports[task_id] == 0:
                            del self.running_imports[task_id]
                            await self.message_bus.emit_event(CoreMessage.GraphMergeCompleted, dict(task_id=task_id))
                if task is not None and not task.done():
                    task.cancel()
                if not result:
                    # make sure the change is aborted in case of transaction
                    log.info(f"Abort update manually: {change_id}")
                    await db.abort_update(change_id)
                await send_to_child(PoisonPill())
                await run_async(updater.join, stale)
                if updater.is_alive():
                    log.warning(f"Process is still alive after poison pill. Terminate process {updater.pid}")
                    with suppress(Exception):
                        updater.terminate()
                    await asyncio.sleep(3)
                if updater.is_alive():
                    log.warning(f"Process is still alive after terminate. Kill process {updater.pid}")
                    with suppress(Exception):
                        updater.kill()
                    await run_async(updater.join)
                if not updater.is_alive():
                    with suppress(Exception):
                        updater.close()



from __future__ import annotations

import asyncio
import hashlib
import json
import logging
import random
import re
import string
import sys
import uuid
from asyncio import Future, AbstractEventLoop, Task
from collections import defaultdict
from collections.abc import Iterable
from contextlib import suppress
from datetime import timedelta, datetime, timezone
from typing import (
    Any,
    Callable,
    Optional,
    TypeVar,
    Mapping,
    AsyncGenerator,
    Dict,
    List,
    Tuple,
    AsyncIterator,
    Iterator,
    Union,
    Sequence,
    Awaitable,
)

from dateutil.parser import isoparse, parse as parse_date

from resotocore.error import RestartService
from resotocore.service import Service
from resotocore.types import JsonElement, Json
from resotolib.durations import parse_duration

log = logging.getLogger(__name__)

AnyT = TypeVar("AnyT")
AnyR = TypeVar("AnyR")


class Periodic(Service):
    """
    Periodic execution of a function based on a defined frequency that can be started and stopped.
    """

    def __init__(
        self,
        name: str,
        func: Callable[[], Any],
        frequency: timedelta,
        first_run: Optional[timedelta] = None,
        loop: Optional[AbstractEventLoop] = None,
    ):
        self.name = name
        self.func = func
        self.frequency = frequency
        self.first_run = first_run if first_run else frequency
        self._task: Optional[Task[None]] = None
        self._loop = loop

    @property
    def started(self) -> bool:
        return self._task is not None

    async def start(self) -> None:
        if self._task is None:
            # Start task to call func periodically:
            self._task = asyncio.ensure_future(self._run(), loop=self._loop)
            log.debug(f"Periodic task {self.name} has been started.")

    async def stop(self) -> None:
        # Stop task and await it stopped:
        if self._task is not None:
            self._task.cancel()
            with suppress(asyncio.CancelledError):
                await self._task

    async def _run(self) -> None:
        await asyncio.sleep(self.first_run.total_seconds())
        while True:
            log.debug(f"Execute periodic task {self.name}.")
            try:
                result = self.func()
                if isinstance(result, Awaitable):
                    await result
            except Exception as ex:
                log.error(
                    f"Periodic function {self.name} caught an exception: {ex}",
                    exc_info=ex,
                )
            await asyncio.sleep(self.frequency.total_seconds())


# noinspection PyUnusedLocal
async def async_noop(*args: Any, **kwargs: Any) -> None:
    pass


def identity(o: AnyT) -> AnyT:
    return o


def count_iterator(start: int = 0) -> Iterator[int]:
    return iter(range(start, sys.maxsize))


def json_hash(js: Json) -> str:
    sha256 = hashlib.sha256()
    sha256.update(json.dumps(js, sort_keys=True).encode("utf-8"))
    return sha256.hexdigest()


def pop_keys(d: Dict[AnyT, AnyR], keys: List[AnyT]) -> Dict[AnyT, AnyR]:
    res = dict(d)
    for key in keys:
        res.pop(key, None)  # type: ignore
    return res


UTC_Date_Format = "%Y-%m-%dT%H:%M:%SZ"

# short iso 8601 format, e.g. 20230504T080910Z
UTC_Date_Format_short = "%Y%m%dT%H%M%SZ"


def rnd_str(str_len: int = 10) -> str:
    return "".join(random.choice(string.ascii_uppercase + string.digits) for _ in range(str_len))


def utc() -> datetime:
    return datetime.now(timezone.utc)


def utc_str(dt: Optional[datetime] = None, date_format: str = UTC_Date_Format) -> str:
    if dt is None:
        dt = utc()
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=timezone.utc)
    elif dt.tzinfo == timezone.utc:
        pass
    else:
        dt = dt.astimezone(timezone.utc)
    return dt.strftime(date_format)


def from_utc(date_string: str) -> datetime:
    return isoparse(date_string)


def parse_utc(date_string: str) -> datetime:
    try:
        dt = datetime.fromisoformat(date_string)
    except Exception:
        dt = parse_date(date_string)
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=timezone.utc)
    elif dt.tzinfo == timezone.utc:
        dt = dt.astimezone(timezone.utc)
    return dt


def duration(d: str) -> timedelta:
    return parse_duration(d)


def uuid_str(from_object: Optional[Any] = None) -> str:
    if from_object:
        return str(uuid.uuid5(uuid.NAMESPACE_DNS, from_object))
    else:
        return str(uuid.uuid1())


def partition_by(f: Callable[[AnyT], bool], iterable: Iterable[AnyT]) -> Tuple[List[AnyT], List[AnyT]]:
    """
    Partition a list based on provided function.
    :param f: the function to test every element.
    :param iterable: the iterable to walk.
    :return: a tuple with two lists. The first list contains all elements for which the function returned True
             the second list contains all elements for which the function returned False.
    """
    left: List[AnyT] = []
    right: List[AnyT] = []
    for p in iterable:
        (left if f(p) else right).append(p)
    return left, right


def group_by(f: Callable[[AnyT], AnyR], iterable: Iterable[AnyT]) -> Dict[AnyR, List[AnyT]]:
    """
    Group iterable by key provided by given key function.
    :param f: the function to be applied on every element that yields the key.
    :param iterable: the iterable to walk.
    :return: a dictionary with the keys provided by the key function and the values from the iterable.
    """
    v = defaultdict(list)
    for item in iterable:
        key = f(item)
        v[key].append(item)
    return v


def non_empty(el: Iterable[Any]) -> bool:
    return bool(el)


def empty(el: Iterable[Any]) -> bool:
    return not non_empty(el)


def combine_optional(
    left: Optional[AnyT], right: Optional[AnyT], combine: Callable[[AnyT, AnyT], AnyT]
) -> Optional[AnyT]:
    if left and right:
        return combine(left, right)
    elif left:
        return left
    else:
        return right


def interleave(elements: List[AnyT]) -> List[Tuple[AnyT, AnyT]]:
    if len(elements) < 2:
        return []
    else:
        nxt = iter(elements)
        next(nxt)
        return list(zip(elements, nxt))


# we expect a callable that returns a truthy value.
# Due to limitations of lambda expressions we use Any here.
def exist(f: Callable[[Any], Any], iterable: Iterable[Any]) -> bool:
    """
    Items are passed to the callable as long as it returns False.
    Return True once the callable finds one True, otherwise return False.
    :param f: the callable that needs to accept items from Iterable.
    :param iterable: walk over this iterable
    :return: True if the item exist, otherwise False.
    """
    for a in iterable:
        if f(a):
            return True
    return False


# we expect a callable that returns a truthy value.
# Due to limitations of lambda expressions we use Any here.
def first(f: Callable[[AnyT], Any], iterable: Iterable[AnyT]) -> Optional[AnyT]:
    for a in iterable:
        if f(a):
            return a
    return None


def if_set(x: Optional[AnyT], func: Callable[[AnyT], Any], if_not: Any = None) -> Any:
    """
    Conditional execute based if the option is defined.
    :param x: the value to check.
    :param func: the function to call if the item is defined.
    :param if_not: the value to return if the item is not defined.
    :return: the result of the function or if_not
    """
    return func(x) if x is not None else if_not


def value_in_path_get(element: JsonElement, path_or_name: Union[List[str], str], if_none: AnyT) -> AnyT:
    result = value_in_path(element, path_or_name)
    return result if result is not None and isinstance(result, type(if_none)) else if_none


def value_in_path(element: JsonElement, path_or_name: Union[List[str], str]) -> Optional[Any]:
    path = path_or_name if isinstance(path_or_name, list) else path_or_name.split(".")
    at = len(path)

    def at_idx(current: JsonElement, idx: int) -> Optional[Any]:
        if at == idx:
            return current
        elif current is None or not isinstance(current, dict) or path[idx] not in current:
            return None
        else:
            return at_idx(current[path[idx]], idx + 1)

    return at_idx(element, 0)


def deep_merge(left: Json, right: Json) -> Json:
    """
    Merge the right json into the left json.
    All values in right will be set on the left side.
    All values not existing in right will be preserved on the left side.
    :return: the deeply merged json object.
    """

    def merge(key: str) -> JsonElement:
        left_value: JsonElement = left.get(key)
        right_value: JsonElement = right.get(key)
        if isinstance(right_value, dict):
            return deep_merge(left_value if isinstance(left_value, dict) else {}, right_value)
        elif key in right:
            return right_value
        else:
            return left_value

    return {k: merge(k) for k in set(left.keys()).union(right.keys())}


def set_value_in_path(element: JsonElement, path_or_name: Union[List[str], str], js: Optional[Json] = None) -> Json:
    path = path_or_name if isinstance(path_or_name, list) else path_or_name.split(".")
    at = len(path) - 1

    def at_idx(current: Json, idx: int) -> None:
        if at == idx:
            current[path[-1]] = element
        else:
            value = current.get(path[idx])
            if not isinstance(value, dict):
                value = {}
                current[path[idx]] = value
            at_idx(value, idx + 1)

    js = js if js is not None else {}
    at_idx(js, 0)
    return js


def del_value_in_path(element: JsonElement, path_or_name: Union[List[str], str]) -> JsonElement:
    path = path_or_name if isinstance(path_or_name, list) else path_or_name.split(".")
    pl = len(path) - 1

    def at_idx(current: JsonElement, idx: int) -> JsonElement:
        if current is None or not isinstance(current, dict) or path[idx] not in current:
            return element
        elif pl == idx:
            current.pop(path[-1], None)
            return element
        else:
            result = at_idx(current[path[idx]], idx + 1)
            if not current[path[idx]]:
                current[path[idx]] = None
            return result

    return at_idx(element, 0)


async def force_gen(gen: AsyncIterator[AnyT]) -> AsyncIterator[AnyT]:
    async def with_first(elem: AnyT) -> AsyncGenerator[AnyT, None]:
        yield elem
        async for a in gen:
            yield a

    try:
        return with_first(await gen.__anext__())  # pylint: disable=unnecessary-dunder-call
    except StopAsyncIteration:
        return gen


def set_future_result(future: Future[Any], result: Any) -> None:
    if not future.done():
        if isinstance(result, Exception):
            future.set_exception(result)
        else:
            future.set_result(result)


def __mute_async_exception_reporting_on_current_loop() -> None:
    # Exceptions happening in the async loop during shutdown.
    def exception_handler(_: Any, context: Any) -> None:
        log.debug(f"Error from async loop during shutdown: {context}")

    log.info("Shutdown initiated for current process.")
    with suppress(Exception):
        loop = asyncio.get_running_loop()
        if loop:
            loop.set_exception_handler(exception_handler)


def restart_service(reason: str) -> None:
    __mute_async_exception_reporting_on_current_loop()
    raise RestartService(reason)


def shutdown_process(exit_code: int) -> None:
    __mute_async_exception_reporting_on_current_loop()
    sys.exit(exit_code)


class AccessNone:
    def __init__(self, not_existent: Any = None):
        self.__not_existent = not_existent

    def __getitem__(self, item: Any) -> Any:
        return self

    def __getattr__(self, name: Any) -> Any:
        return self

    def __str__(self) -> str:
        return str(self.__not_existent)

    def __eq__(self, other: Any) -> bool:
        return other is None or isinstance(other, AccessNone)

    def __iter__(self) -> Iterator[Any]:
        return self

    def __next__(self) -> None:
        raise StopIteration()

    @property
    def is_none(self) -> bool:
        return True


class AccessJsonList(List[Any]):
    def __init__(self, seq: Iterable[Any], not_existent: Any = None):
        super().__init__(seq)
        self.__not_existent = AccessNone(not_existent)

    def __getitem__(self, item: Any) -> Any:
        try:
            return super().__getitem__(item)
        except Exception:
            return AccessNone(self.__not_existent)


class AccessJson(Dict[Any, Any]):
    """
    Extend dict in order to allow python like property access
    as well as exception safe access for non existent properties.
    """

    def __init__(
        self,
        mapping: Mapping[Any, Any],
        not_existent: Any = None,
        simple_formatter: Callable[[Any], Any] = identity,
        self_name: Optional[str] = None,  # the whole object is returned if this property name is requested.
    ) -> None:
        super().__init__(mapping)
        self.__not_existent = AccessNone(not_existent)
        self.__simple_formatter = simple_formatter
        self.__self_name = self_name

    def __getitem__(self, item: Any) -> Any:
        if item in self:
            iv = super().__getitem__(item)
            return (
                AccessJson.wrap(iv, self.__not_existent, self.__simple_formatter)
                if iv is not None
                else self.__not_existent
            )
        elif self.__self_name is not None and item == self.__self_name:
            return self
        else:
            return self.__not_existent

    def __getattr__(self, name: Any) -> Any:
        return self.__getitem__(name)

    def __str__(self) -> str:
        return json.dumps(self)

    @property
    def is_none(self) -> bool:
        return False

    @staticmethod
    def wrap(
        obj: Any,
        not_existent: Any = None,
        simple_formatter: Callable[[Any], Any] = identity,
        self_name: Optional[str] = None,
    ) -> Any:
        if isinstance(obj, (str, int, float, AccessJson)):
            return simple_formatter(obj)
        # dict like data structure -> wrap whole element
        elif isinstance(obj, Mapping):
            return AccessJson(obj, not_existent, simple_formatter, self_name)
        # list like data structure -> wrap all elements
        elif isinstance(obj, Sequence):
            return AccessJsonList(
                (AccessJson.wrap(item, not_existent, simple_formatter, self_name) for item in obj), not_existent
            )
        # simply return the object
        else:
            return obj

    @staticmethod
    def wrap_list(
        obj: Any, not_existent: Any = AccessNone(None), simple_formatter: Callable[[Any], Any] = identity
    ) -> List[AccessJson]:
        # only here for a typed result
        return AccessJson.wrap(obj, not_existent, simple_formatter)  # type: ignore

    @staticmethod
    def wrap_object(
        obj: Any, not_existent: Any = AccessNone(None), simple_formatter: Callable[[Any], Any] = identity
    ) -> AccessJson:
        # only here for a typed result
        return AccessJson.wrap(obj, not_existent, simple_formatter)  # type: ignore


def check_graph_name(name: str) -> None:
    if not re.match(r"^[a-zA-Z0-9_]+$", name):
        raise ValueError(f"Graph name {name} contains invalid characters, only alphanumeric characters are allowed.")



from __future__ import annotations

import logging
from datetime import timedelta
from pathlib import Path
from ssl import SSLContext, create_default_context, Purpose
from tempfile import TemporaryDirectory
from typing import Tuple, Optional, List, Union, Dict

from arango.database import StandardDatabase
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.x509 import Certificate, CertificateSigningRequest

from resotocore.core_config import CoreConfig, CertificateConfig
from resotocore.service import Service
from resotocore.types import Json
from resotocore.util import Periodic
from resotolib.utils import get_local_ip_addresses, get_local_hostnames
from resotolib.x509 import (
    bootstrap_ca,
    gen_rsa_key,
    gen_csr,
    sign_csr,
    cert_to_bytes,
    key_to_bytes,
    load_key_from_bytes,
    load_cert_from_bytes,
    cert_fingerprint,
    load_csr_from_bytes,
    write_key_to_file,
    write_cert_to_file,
    load_key_from_file,
    load_cert_from_file,
    write_ca_bundle,
)

log = logging.getLogger(__name__)


class CertificateHandler(Service):
    def __init__(self, config: CoreConfig, ca_key: RSAPrivateKey, ca_cert: Certificate, temp_dir: Path) -> None:
        self.config = config
        self._ca_key = ca_key
        self._ca_cert = ca_cert
        self._ca_cert_bytes = cert_to_bytes(ca_cert)
        self._ca_cert_fingerprint = cert_fingerprint(ca_cert)
        self._host_key, self._host_cert = self.__create_host_certificate(config.api.host_certificate, ca_key, ca_cert)
        self._host_context = self._create_host_context(config, self._host_cert, self._host_key)
        self._client_context = self.__create_client_context(config, ca_cert)
        self._ca_bundle = temp_dir / "ca-bundle.crt"
        self.__recreate_ca_file()  # write the CA bundle to the temp dir
        self._ca_cert_recreate = Periodic("recreate ca bundle file", self.__recreate_ca_file, timedelta(hours=1))

    async def start(self) -> None:
        await self._ca_cert_recreate.start()

    async def stop(self) -> None:
        await self._ca_cert_recreate.stop()

    def __recreate_ca_file(self) -> None:
        write_ca_bundle(self._ca_cert, str(self._ca_bundle))

    @property
    def ca_cert(self) -> Certificate:
        return self._ca_cert

    @property
    def ca_bundle(self) -> Path:
        return self._ca_bundle

    @property
    def authority_certificate(self) -> Tuple[bytes, str]:
        return self._ca_cert_bytes, self._ca_cert_fingerprint

    @property
    def host_certificate(self) -> Tuple[RSAPrivateKey, Certificate]:
        return self._host_key, self._host_cert

    def create_key_and_cert(
        self, common_name: str, dns_names: List[str], ip_addresses: List[str], days_valid: int
    ) -> Tuple[RSAPrivateKey, Certificate]:
        key = gen_rsa_key()
        csr = gen_csr(
            key,
            include_loopback=False,
            common_name=common_name,
            san_dns_names=dns_names,
            san_ip_addresses=ip_addresses,
            discover_local_dns_names=False,
            discover_local_ip_addresses=False,
        )
        cert = sign_csr(csr, self._ca_key, self._ca_cert, days_valid)
        return key, cert

    def sign(
        self,
        csr_or_bytes: Union[CertificateSigningRequest, bytes],
        days_valid: int = 365,
        server_auth: bool = True,
        client_auth: bool = True,
        key_usage: Optional[Dict[str, bool]] = None,
    ) -> Tuple[Certificate, str]:
        csr = load_csr_from_bytes(csr_or_bytes) if isinstance(csr_or_bytes, bytes) else csr_or_bytes
        certificate = sign_csr(csr, self._ca_key, self._ca_cert, days_valid, server_auth, client_auth, key_usage)
        return certificate, cert_fingerprint(certificate)

    @property
    def host_context(self) -> Optional[SSLContext]:
        return self._host_context

    @property
    def client_context(self) -> SSLContext:
        return self._client_context

    @staticmethod
    def __create_host_certificate(
        cfg: CertificateConfig, ca_key: RSAPrivateKey, ca_cert: Certificate
    ) -> Tuple[RSAPrivateKey, Certificate]:
        key = gen_rsa_key()
        host_names = get_local_hostnames(
            include_loopback=cfg.include_loopback,
            san_ip_addresses=cfg.san_ip_addresses,
            san_dns_names=cfg.san_dns_names,
        )
        host_ips = get_local_ip_addresses(include_loopback=cfg.include_loopback, san_ip_addresses=cfg.san_ip_addresses)
        log.info(f'Create host certificate for hostnames:{", ".join(host_names)} and ips:{", ".join(host_ips)}')
        csr = gen_csr(
            key,
            common_name=cfg.common_name,
            san_dns_names=list(host_names),
            san_ip_addresses=list(host_ips),
            include_loopback=cfg.include_loopback,
        )
        cert = sign_csr(csr, ca_key, ca_cert)
        return key, cert

    @staticmethod
    def _create_host_context(config: CoreConfig, host_cert: Certificate, host_key: RSAPrivateKey) -> SSLContext:
        args = config.args
        # noinspection PyTypeChecker
        ctx = create_default_context(purpose=Purpose.CLIENT_AUTH)
        if config.args.cert:
            log.info("Using TLS certificate from command line.")
            # Use the certificate provided via cmd line flags
            ctx.load_cert_chain(args.cert, args.cert_key, args.cert_key_pass)
        else:
            log.info("Using TLS certificate from data store.")
            # ssl library wants to load cert/key from file: put it into a temp directory for loading
            with TemporaryDirectory() as td:
                cert_file = Path(td, "cert")
                key_file = Path(td, "key")
                write_cert_to_file(host_cert, str(cert_file))
                write_key_to_file(host_key, str(key_file))
                ctx.load_cert_chain(str(cert_file), str(key_file), args.ca_cert_key_pass)
        return ctx

    @staticmethod
    def __create_client_context(config: CoreConfig, ca_cert: Certificate) -> SSLContext:
        # noinspection PyTypeChecker
        ctx = create_default_context(purpose=Purpose.SERVER_AUTH)
        if config.args.ca_cert:
            ctx.load_verify_locations(cafile=config.args.ca_cert)
        else:
            ca_bytes = cert_to_bytes(ca_cert).decode("utf-8")
            ctx.load_verify_locations(cadata=ca_bytes)
        return ctx

    @staticmethod
    def lookup(config: CoreConfig, db: StandardDatabase, temp_dir: Path) -> CertificateHandler:
        args = config.args
        # if we get a ca certificate from the command line, use it
        if args.ca_cert and args.ca_cert_key:
            ca_key = load_key_from_file(args.ca_cert_key, args.ca_cert_key_pass)
            ca_cert = load_cert_from_file(args.ca_cert_cert)
            log.info(f"Using CA certificate from command line. fingerprint:{cert_fingerprint(ca_cert)}")
            return CertificateHandler(config, ca_key, ca_cert, temp_dir)

        # otherwise, load from database or create it
        sd = db.collection("system_data")
        maybe_ca: Optional[Json] = sd.get("ca")  # type: ignore
        if maybe_ca and isinstance(maybe_ca.get("key"), str) and isinstance(maybe_ca.get("certificate"), str):
            log.debug("Found existing certificate in data store.")
            key = load_key_from_bytes(maybe_ca["key"].encode("utf-8"))
            certificate = load_cert_from_bytes(maybe_ca["certificate"].encode("utf-8"))
            log.info(f"Using CA certificate from database. fingerprint:{cert_fingerprint(certificate)}")
            return CertificateHandler(config, key, certificate, temp_dir)
        else:
            wo = "with" if args.ca_cert_key_pass else "without"
            key, certificate = bootstrap_ca()
            log.info(
                f"No ca certificate found - create new one {wo} passphrase. fingerprint:{cert_fingerprint(certificate)}"
            )
            key_string = key_to_bytes(key, args.ca_cert_key_pass).decode("utf-8")
            certificate_string = cert_to_bytes(certificate).decode("utf-8")
            sd.insert({"_key": "ca", "key": key_string, "certificate": certificate_string})
            return CertificateHandler(config, key, certificate, temp_dir)



import asyncio

from datetime import timedelta
from typing import Optional, AsyncIterator, List, Dict, Any, cast
import attrs
import yaml
import os
from deepdiff import DeepDiff
import hashlib

from resotocore.analytics import AnalyticsEventSender, CoreEvent
from resotocore.config import ConfigHandler, ConfigEntity, ConfigValidation, ConfigOverride
from resotocore.db.configdb import ConfigEntityDb, ConfigValidationEntityDb
from resotocore.db.modeldb import ModelDb
from resotocore.message_bus import MessageBus, CoreMessage
from resotocore.model.model import Model, Kind, ComplexKind
from resotocore.service import Service
from resotocore.types import Json, JsonElement
from resotocore.util import uuid_str, deep_merge, first
from resotocore.worker_task_queue import WorkerTaskQueue, WorkerTask, WorkerTaskName
from resotocore.ids import TaskId, ConfigId
from resotocore.core_config import CoreConfig
from resotolib.utils import merge_json_elements

from resotolib.utils import replace_env_vars


class ConfigHandlerService(ConfigHandler, Service):
    def __init__(
        self,
        cfg_db: ConfigEntityDb,
        validation_db: ConfigValidationEntityDb,
        model_db: ModelDb,
        task_queue: WorkerTaskQueue,
        message_bus: MessageBus,
        event_sender: AnalyticsEventSender,
        core_config: CoreConfig,
        override_service: ConfigOverride,
    ) -> None:
        self.cfg_db = cfg_db
        self.validation_db = validation_db
        self.model_db = model_db
        self.task_queue = task_queue
        self.message_bus = message_bus
        self.event_sender = event_sender
        self.core_config = core_config
        self.override_service = override_service
        self.old_overrides: Dict[ConfigId, Json] = {}

    async def coerce_and_check_model(self, cfg_id: ConfigId, config: Json, validate: bool = True) -> Json:
        model = await self.get_configs_model()
        # filter config roots to match top level entries in config
        config_roots = {kind.fqn: kind for kind in model.complex_kinds() if kind.aggregate_root}

        final_config = {}
        if validate:
            for key, value in config.items():
                if key in config_roots:
                    try:
                        value_kind = config_roots[key]
                        coerced = value_kind.check_valid(value, normalize=False, config_context=True)
                        final_config[key] = value_kind.sort_json(coerced or value)
                    except Exception as ex:
                        raise AttributeError(f"Error validating section {key}: {ex}") from ex
                else:
                    final_config[key] = value
        else:
            final_config = config

        # If an external entity needs to approve this change.
        # Method throws if config is not valid according to external approval.
        keys = {key async for key in self.validation_db.keys()}
        parts = cfg_id.split(".")
        # A config with key foo.bla.bar can be validated by foo.bla.bar, foo.bla and foo
        # The longest key is the most specific validator and is used.
        validator = first(lambda x: x in keys, (".".join(parts[0:i]) for i in range(len(parts), 0, -1)))
        if validator:
            validation = await self.validation_db.get(validator)
            if validation and validation.external_validation and validate:
                await self.acknowledge_config_change(validator, final_config)

        # If we come here, everything is fine
        return final_config

    async def coerce_config(self, config: Json) -> Json:
        model = await self.get_configs_model()

        final_config = {}
        for key, value in config.items():
            if key in model:
                value_kind = model[key]
                coerced = value_kind.coerce(value)
                sorted_conf = value_kind.sort_json(coerced) if isinstance(coerced, dict) else coerced
                final_config[key] = sorted_conf
            else:
                final_config[key] = value
        return final_config

    def list_config_ids(self) -> AsyncIterator[ConfigId]:
        return self.cfg_db.keys()

    async def get_config(
        self, cfg_id: ConfigId, apply_overrides: bool = True, resolve_env_vars: bool = True
    ) -> Optional[ConfigEntity]:
        conf = await self.cfg_db.get(cfg_id)
        if conf is None:
            return None

        # apply overrides if they exist and we do not opt out
        # we do not want to apply overrides if the config is to be shown during editing
        overrides = self.override_service.get_override(cfg_id)
        updated_conf = cast(
            Json, merge_json_elements(conf.config, overrides) if overrides and apply_overrides else conf.config
        )

        # reslove env vars
        # we do not want to resolve env vars if the config is to be shown to the user when editing,
        # otherwise sensitive data might be exposed
        if resolve_env_vars:
            resolved_conf = {k: replace_env_vars(v, os.environ) for k, v in updated_conf.items()}
            coerced_conf = await self.coerce_config(resolved_conf)
            updated_conf = coerced_conf

        return attrs.evolve(conf, config=updated_conf)

    async def put_config(self, cfg: ConfigEntity, *, validate: bool = True, dry_run: bool = False) -> ConfigEntity:
        coerced = await self.coerce_and_check_model(cfg.id, cfg.config, validate)
        existing = await self.cfg_db.get(cfg.id)
        if not dry_run and (not existing or existing.config != cfg.config):
            result = await self.cfg_db.update(ConfigEntity(cfg.id, coerced, cfg.revision))
            await self.message_bus.emit_event(CoreMessage.ConfigUpdated, dict(id=result.id, revision=result.revision))
            await self.event_sender.core_event(CoreEvent.SystemConfigurationChanged, result.analytics())
            return result
        else:
            return cfg

    async def patch_config(self, cfg: ConfigEntity, *, validate: bool = True, dry_run: bool = False) -> ConfigEntity:
        current = await self.cfg_db.get(cfg.id)
        current_config = current.config if current else {}
        coerced = await self.coerce_and_check_model(cfg.id, deep_merge(current_config, cfg.config), validate)
        if not dry_run and (not current or current_config != coerced):
            result = await self.cfg_db.update(ConfigEntity(cfg.id, coerced, current.revision if current else None))
            await self.message_bus.emit_event(CoreMessage.ConfigUpdated, dict(id=result.id, revision=result.revision))
            await self.event_sender.core_event(CoreEvent.SystemConfigurationChanged, result.analytics())
            return result
        else:
            return cfg

    async def delete_config(self, cfg_id: ConfigId) -> None:
        await self.cfg_db.delete(cfg_id)
        await self.validation_db.delete(cfg_id)
        await self.message_bus.emit_event(CoreMessage.ConfigDeleted, dict(id=cfg_id))
        await self.event_sender.core_event(CoreEvent.SystemConfigurationDeleted)

    async def copy_config(self, from_cfg_id: ConfigId, to_cfg_id: ConfigId) -> Optional[ConfigEntity]:
        old = await self.cfg_db.get(from_cfg_id)
        if old is None:
            return None
        if await self.cfg_db.get(to_cfg_id) is not None:
            raise ValueError(f"Config with id {to_cfg_id} already exists")
        result = await self.cfg_db.update(ConfigEntity(to_cfg_id, old.config, old.revision))
        await self.message_bus.emit_event(
            CoreMessage.ConfigUpdated, dict(old=old.id, new=result.id, revision=result.revision)
        )
        await self.event_sender.core_event(CoreEvent.SystemConfigurationChanged, result.analytics())
        return result

    def list_config_validation_ids(self) -> AsyncIterator[str]:
        return self.validation_db.keys()

    async def get_config_validation(self, cfg_id: str) -> Optional[ConfigValidation]:
        return await self.validation_db.get(cfg_id)

    async def put_config_validation(self, validation: ConfigValidation) -> ConfigValidation:
        return await self.validation_db.update(validation)

    async def get_configs_model(self) -> Model:
        kinds = [kind async for kind in self.model_db.all()]
        return Model.from_kinds(list(kinds))

    async def update_configs_model(self, kinds: List[Kind]) -> Model:
        # load existing model
        model = await self.get_configs_model()
        # make sure the update is valid, but ignore overlapping property paths, so the same name can
        # have different types in different sections
        updated = model.update_kinds(kinds, check_overlap=False)
        # store all updated kinds
        await self.model_db.update_many(kinds)
        return updated

    async def config_yaml(self, cfg_id: ConfigId, revision: bool = False) -> Optional[str]:
        config = await self.get_config(cfg_id, apply_overrides=False, resolve_env_vars=False)
        if config:
            model = await self.get_configs_model()

            # returns the overridden config with comments about the changes
            def overridden_parts(existing: JsonElement, update: JsonElement) -> JsonElement:
                if isinstance(update, dict):
                    return {
                        key: overridden_parts(
                            existing.get(key) if isinstance(existing, dict) else existing, update[key]
                        )
                        for key in set(update.keys())
                    }
                else:

                    def mkstr(val: Any) -> str:
                        if isinstance(val, list):
                            return f'[{", ".join(val)}]'
                        return str(val)

                    return mkstr(update)

            yaml_str = ""

            overrides = overridden_parts(config.config, self.override_service.get_override(cfg_id) or {})
            overrides_json = overrides if isinstance(overrides, dict) else {}

            for num, (key, value) in enumerate(config.config.items()):
                maybe_kind = model.get(key)
                if isinstance(maybe_kind, ComplexKind):
                    part = maybe_kind.create_yaml(value, initial_level=1, overrides=overrides_json.get(key) or {})
                    if num > 0:
                        yaml_str += "\n"
                    yaml_str += key + ":" + part
                else:
                    yaml_str += yaml.dump({key: value}, sort_keys=False, allow_unicode=True)

            # mix the revision into the yaml document
            if revision and config.revision:
                yaml_str += (
                    "\n\n# This property is not part of the configuration but defines the revision "
                    "of this document.\n# Please leave it here to avoid conflicting writes.\n"
                    f'_revision: "{config.revision}"'
                )

            return yaml_str
        else:
            return None

    async def acknowledge_config_change(self, cfg_id: str, config: Json) -> None:
        """
        In case an external entity should acknowledge this config change.
        This method either return, which signals success or throws an exception.
        """
        future = asyncio.get_event_loop().create_future()
        task = WorkerTask(
            TaskId(uuid_str()),
            WorkerTaskName.validate_config,
            {"config_id": cfg_id},
            {"task": WorkerTaskName.validate_config, "config": config},
            future,
            timedelta(seconds=30),
        )
        # add task to queue - do not retry
        await self.task_queue.add_task(task)
        # In case the config is not valid or no worker is available
        # this future will throw an exception.
        # Do not handle it here and let the error bubble up.
        await future

    async def start(self) -> None:
        async def on_override_change(new_overrides: Dict[ConfigId, Json]) -> None:
            def updated_revision(conf: Json, override: Json) -> str:
                m = hashlib.sha1(usedforsecurity=False)
                m.update(yaml.dump(conf).encode("utf-8"))
                m.update(yaml.dump(override).encode("utf-8"))
                return m.hexdigest()

            diff = DeepDiff(self.old_overrides, new_overrides, ignore_order=True)
            if diff.affected_root_keys:
                affected_configs = diff.affected_root_keys

                for config_id in affected_configs:
                    # the new and updated version, since the override is already applied
                    config = await self.get_config(config_id)
                    if config:
                        new_revision = updated_revision(config.config, new_overrides.get(config_id) or {})
                        await self.message_bus.emit_event(
                            CoreMessage.ConfigUpdated, dict(id=config.id, revision=new_revision)
                        )
                        await self.event_sender.core_event(CoreEvent.SystemConfigurationChanged, config.analytics())

                self.old_overrides = new_overrides

        self.old_overrides = self.override_service.get_all_overrides()
        self.override_service.add_override_change_hook(on_override_change)



from __future__ import annotations

import abc
from collections import defaultdict
from datetime import datetime, timedelta

from attrs import define, field, evolve
from functools import reduce, partial, cached_property
from itertools import chain
from typing import Mapping, Union, Optional, Any, ClassVar, Dict, List, Tuple, Callable, Set, Iterable

from jsons import set_deserializer

from resotocore.model.graph_access import EdgeTypes, Direction
from resotocore.model.resolve_in_graph import GraphResolver
from resotocore.model.typed_model import to_js_str
from resotocore.types import Json, JsonElement, EdgeType
from resotocore.util import combine_optional, utc_str, utc
from resotolib.durations import duration_str

PathRoot = "/"


def variable_to_absolute(section: Optional[str], name: str) -> str:
    if name.startswith(PathRoot):
        return name[1:]
    elif section and section != PathRoot:
        return section + "." + name
    else:
        return name


def variable_to_relative(section: str, name: str) -> str:
    if name.startswith(PathRoot):
        return name
    elif name.startswith(f"{section}."):
        return name[len(section) + 1 :]  # noqa: E203a
    else:
        return PathRoot + name


def is_ancestor_descendant(name: str) -> bool:
    return name not in GraphResolver.resolved_property_names and (
        name.startswith("ancestors.") or name.startswith("descendants.")
    )


@define(order=True, hash=True, frozen=True)
class Template:
    """
    A template has a name and a template string.
    The template string might contain placeholder values.
    """

    name: str  # the name of the template
    template: str  # the template string with placeholders


@define(order=True, hash=True, frozen=True)
class Expandable:
    """
    An expandable refers to a template with a given name
    and has properties to render this template to a final string.
    """

    template: str  # the name of the template
    props: Json  # the properties to render this template


class P:
    def __init__(self, name: str, **kwargs: Any):
        self.name = name
        self.args = kwargs

    @staticmethod
    def single(name: str) -> P:
        return P(name)

    @staticmethod
    def array(name: str) -> PArray:
        return PArray(name)

    @staticmethod
    def with_id(uid: str) -> Term:
        return IdTerm(uid)

    @staticmethod
    def of_kind(name: str) -> Term:
        return IsTerm([name])

    @staticmethod
    def all() -> Term:
        return AllTerm()

    @staticmethod
    def context(name: str, *terms: Term) -> Term:
        return ContextTerm(name, reduce(lambda a, b: a & b, terms, AllTerm()))  # type: ignore

    @staticmethod
    def function(fn: str) -> PFunction:
        return PFunction(fn)

    def __gt__(self, other: Any) -> Predicate:
        return self.gt(other)

    def __ge__(self, other: Any) -> Predicate:
        return self.ge(other)

    def __lt__(self, other: Any) -> Predicate:
        return self.lt(other)

    def __le__(self, other: Any) -> Predicate:
        return self.le(other)

    def __eq__(self, other: Any) -> Predicate:  # type: ignore
        return self.eq(other)

    def __ne__(self, other: Any) -> Predicate:  # type: ignore
        return self.ne(other)

    def gt(self, other: Any) -> Predicate:
        return Predicate(self.name, ">", other, self.args)

    def ge(self, other: Any) -> Predicate:
        return Predicate(self.name, ">=", other, self.args)

    def lt(self, other: Any) -> Predicate:
        return Predicate(self.name, "<", other, self.args)

    def le(self, other: Any) -> Predicate:
        return Predicate(self.name, "<=", other, self.args)

    def eq(self, other: Any) -> Predicate:
        return Predicate(self.name, "==", other, self.args)

    def ne(self, other: Any) -> Predicate:
        return Predicate(self.name, "!=", other, self.args)

    def matches(self, regex: str) -> Predicate:
        return Predicate(self.name, "=~", regex, self.args)

    def not_matches(self, regex: str) -> Predicate:
        return Predicate(self.name, "!~", regex, self.args)

    def is_in(self, other: List[Any]) -> Predicate:
        return Predicate(self.name, "in", other, self.args)

    def is_not_in(self, other: List[Any]) -> Predicate:
        return Predicate(self.name, "not in", other, self.args)


class PFunction:
    def __init__(self, fn: str):
        self.fn = fn

    def on(self, name: str, *args: Any) -> FunctionTerm:
        return FunctionTerm(self.fn, name, list(args))


class PArray:
    def __init__(self, name: str):
        self.name = name

    def for_any(self) -> P:
        return P(self.name, filter="any")

    def for_none(self) -> P:
        return P(self.name, filter="none")

    def for_all(self) -> P:
        return P(self.name, filter="all")


@define(order=True, hash=True, frozen=True)
class Term(abc.ABC):
    """
    @startuml
    class Term
    Term <|-- CombinedTerm
    Term <|-- NotTerm
    Term <|-- MergeTerm
    Term <|--- AllTerm
    Term <|--- FulltextTerm
    Term <|--- Predicate
    Term <|--- ContextTerm
    Term <|--- IdTerm
    Term <|--- IsTerm
    Term <|--- FunctionTerm
    NotTerm --> Term: not
    CombinedTerm --> Term: left
    CombinedTerm --> Term: right
    MergeTerm --> Term: pre_filter
    MergeTerm --> Term: post_filter
    ContextTerm *--> Term: term
    @enduml
    """

    def __or__(self, other: Term) -> Term:
        return self.or_term(other)

    def __and__(self, other: Term) -> Term:
        return self.and_term(other)

    def not_term(self) -> NotTerm:
        return NotTerm(self)

    def combine(self, op: str, other: Term) -> Term:
        if op == "or":
            return self.or_term(other)
        elif op == "and":
            return self.and_term(other)
        else:
            raise AttributeError(f"Don't know how to combine with {op}")

    @property
    def is_all(self) -> bool:
        return isinstance(self, AllTerm)

    def or_term(self, other: Term) -> Term:
        if isinstance(self, AllTerm):  # all or x == all
            return self
        elif isinstance(other, AllTerm):  # x or all == all
            return other
        elif isinstance(self, MergeTerm):  # combining a merge term needs special handling
            return self.or_merge_term(other)
        elif isinstance(other, MergeTerm):  # combining a merge term needs special handling
            return other.or_merge_term(self)
        else:
            return CombinedTerm(self, "or", other)

    def and_term(self, other: Term) -> Term:
        if isinstance(self, AllTerm):  # all and x == x
            return other
        elif isinstance(other, AllTerm):  # x and all == x
            return self
        elif isinstance(self, MergeTerm):  # combining a merge term needs special handling
            return self.and_merge_term(other)
        elif isinstance(other, MergeTerm):  # combining a merge term needs special handling
            return other.and_merge_term(self)
        else:
            return CombinedTerm(self, "and", other)

    def change_variable(self, fn: Callable[[str], str]) -> Term:
        def walk(term: Term) -> Term:
            if isinstance(term, CombinedTerm):
                return CombinedTerm(walk(term.left), term.op, walk(term.right))
            if isinstance(term, ContextTerm):
                return ContextTerm(fn(term.name), term.term)
            elif isinstance(term, Predicate):
                return Predicate(fn(term.name), term.op, term.value, term.args)
            elif isinstance(term, FunctionTerm):
                return FunctionTerm(term.fn, fn(term.property_path), term.args)
            elif isinstance(term, MergeTerm):
                post = walk(term.post_filter) if term.post_filter else None
                return MergeTerm(walk(term.pre_filter), [mq.change_variable(fn) for mq in term.merge], post)
            elif isinstance(term, NotTerm):
                return NotTerm(walk(term.term))
            else:
                return term

        return walk(self)

    def find_term(self, fn: Callable[[Term], bool]) -> Optional[Term]:
        if fn(self):
            return self
        if isinstance(self, CombinedTerm):
            return self.left.find_term(fn) or self.right.find_term(fn)
        elif isinstance(self, NotTerm):
            return self.term.find_term(fn)
        elif isinstance(self, ContextTerm):
            return self.term.find_term(fn)
        elif isinstance(self, MergeTerm):

            def walk_merge_queries(mt: MergeTerm) -> Optional[Term]:
                for mq in mt.merge:
                    for p in mq.query.parts:
                        if (term := p.term.find_term(fn)) is not None:
                            return term
                return None

            return (
                self.pre_filter.find_term(fn)
                or (self.post_filter.find_term(fn) if self.post_filter else None)
                or walk_merge_queries(self)
            )
        else:
            return None

    def find_terms(self, fn: Callable[[Term], bool], **kwargs: bool) -> List[Term]:
        if fn(self):
            return [self]
        elif isinstance(self, CombinedTerm):
            return self.left.find_terms(fn, **kwargs) + self.right.find_terms(fn, **kwargs)
        elif isinstance(self, ContextTerm) and kwargs.get("in_context_term", True):
            return self.term.find_terms(fn, **kwargs)
        elif isinstance(self, NotTerm):
            return self.term.find_terms(fn, **kwargs)
        elif isinstance(self, MergeTerm):
            result = self.pre_filter.find_terms(fn, **kwargs)
            if self.post_filter:
                result.extend(self.post_filter.find_terms(fn, **kwargs))
            result.extend(r for mq in self.merge for p in mq.query.parts for r in p.term.find_terms(fn, **kwargs))
            return result
        else:
            return []

    def contains_term_type(self, clazz: type) -> bool:
        return self.find_term(lambda x: isinstance(x, clazz)) is not None

    def split_term_by(self, check_fn: Callable[[Term], bool]) -> Tuple[Term, Term]:
        before_merge: Term = AllTerm()
        after_merge: Term = AllTerm()

        def walk_term(term: Term) -> None:
            # precondition: this method is only called with a term that has ancestor/descendant
            nonlocal before_merge
            nonlocal after_merge
            if isinstance(term, CombinedTerm):
                left_has_ad = check_fn(term.left)
                right_has_ad = check_fn(term.right)
                if term.op == "or":
                    after_merge = after_merge & term
                elif left_has_ad and right_has_ad:
                    walk_term(term.left)
                    walk_term(term.right)
                elif left_has_ad:
                    before_merge = before_merge & term.right
                    walk_term(term.left)
                elif right_has_ad:
                    before_merge = before_merge & term.left
                    walk_term(term.right)
                else:
                    raise NotImplementedError("Logic unsound. This case should not happen!")
            elif isinstance(term, MergeTerm):
                # in case pre- and post- filter are defined, handle it as AND combined term
                # background: pre- and post- filter will be applied on the result
                #             that effectively reflects an and combination.
                #             The merge part only merges data to the existing values.
                if term.post_filter:
                    walk_term(CombinedTerm(term.pre_filter, "and", term.post_filter))
                else:
                    walk_term(term.pre_filter)
            else:
                after_merge = after_merge & term

        if check_fn(self):
            walk_term(self)
            return before_merge, after_merge
        else:
            return self, AllTerm()

    def split_by_usage(self) -> Tuple[Term, Term]:
        return self.split_term_by(
            lambda x: x.find_term(lambda t: isinstance(t, Predicate) and t.name.startswith("usage")) is not None
        )

    # noinspection PyUnusedLocal
    @staticmethod
    def from_json(js: Dict[str, Any], _: type = object, **kwargs: Any) -> Term:
        if isinstance(js.get("left"), dict) and isinstance(js.get("right"), dict) and isinstance(js.get("op"), str):
            left = Term.from_json(js["left"])
            right = Term.from_json(js["right"])
            return CombinedTerm(left, js["op"], right)
        elif isinstance(js.get("name"), str) and isinstance(js.get("op"), str):
            args = js["args"] if isinstance(js.get("args"), dict) else {}
            return Predicate(js["name"], js["op"], js["value"], args)
        elif isinstance(js.get("name"), str) and isinstance(js.get("predicates"), list):
            return ContextTerm(js["name"], js["predicates"])
        elif isinstance(js.get("fn"), str) and isinstance(js.get("property_path"), str):
            argv: list = js["args"] if isinstance(js.get("args"), list) else []  # type: ignore
            return FunctionTerm(js["fn"], js["property_path"], argv)
        elif isinstance(js.get("kind"), str):
            return IsTerm(js["kind"])
        elif isinstance(js.get("id"), str):
            return IdTerm(js.get("id"))  # type: ignore
        else:
            raise AttributeError(f"Can not parse json into query: {js}")


@define(order=True, hash=True, frozen=True)
class AllTerm(Term):
    _instance = None

    def __new__(cls) -> AllTerm:
        if cls._instance is None:
            cls._instance = super(AllTerm, cls).__new__(cls)
        return cls._instance

    def __str__(self) -> str:
        return "all"


@define(order=True, hash=True, frozen=True)
class NotTerm(Term):
    term: Term

    def __str__(self) -> str:
        return f"not({self.term})"


@define(order=True, hash=True, frozen=True)
class FulltextTerm(Term):
    text: str

    def __str__(self) -> str:
        return f'"{self.text}"'


@define(order=True, hash=True, frozen=True)
class Predicate(Term):
    name: str
    op: str
    value: JsonElement
    args: Mapping[str, JsonElement]

    def __str__(self) -> str:
        modifier = f'{self.args["filter"]} ' if "filter" in self.args else ""
        return f"{self.name} {modifier}{self.op} {self.value_str_rep(self.value)}"

    @staticmethod
    def value_str_rep(value: Any) -> str:
        """
        This method is used to get a string representation of a value.
        :param value: the value to be represented.
        :return: the string representation.
        """
        return to_js_str(value)


@define(order=True, hash=True, frozen=True)
class ContextTerm(Term):
    name: str
    term: Term

    def __str__(self) -> str:
        return f"{self.name}.{{{str(self.term)}}}"

    def visible_predicates(self) -> List[Predicate]:
        """
        This method is not used to render a query, but to get a list of predicates that are visible in the query.
        All predicates have the complete path without any contextual information.
        Idea: a.b[*].{ c=1 and d=2 and e[*].{f=1}} should return [a.b[*].c = 1, a.b[*].d = 2, a.b[*].e[*].f = 1]
        """

        def with_context(path: str, ctx: ContextTerm) -> List[Predicate]:
            result: List[Predicate] = []
            for p in ctx.term.find_terms(lambda x: isinstance(x, Predicate), in_context_term=False):
                result.append(evolve(p, name=f"{path}.{p.name}" if path else p.name))  # type: ignore
            for c in ctx.term.find_terms(lambda x: isinstance(x, ContextTerm)):
                result.extend(with_context(f"{path}.{c.name}" if path else c.name, c))  # type: ignore
            return result

        return with_context(self.name, self)


@define(order=True, hash=True, frozen=True)
class CombinedTerm(Term):
    left: Term
    op: str
    right: Term

    def __str__(self) -> str:
        return f"({self.left} {self.op} {self.right})"


@define(order=True, hash=True, frozen=True)
class IdTerm(Term):
    id: str

    def __str__(self) -> str:
        return f'id("{self.id}")'


@define(order=True, hash=True, frozen=True)
class IsTerm(Term):
    kinds: List[str]

    def __str__(self) -> str:
        kind_string = ", ".join(f'"{a}"' for a in self.kinds)
        kinds = kind_string if len(self.kinds) == 1 else f"[{kind_string}]"
        return f"is({kinds})"


@define(order=True, hash=True, frozen=True)
class FunctionTerm(Term):
    fn: str
    property_path: str
    args: List[Any]

    def __str__(self) -> str:
        args = ", ".join((Predicate.value_str_rep(a) for a in self.args))
        sep = ", " if args else ""
        return f"{self.fn}({self.property_path}{sep}{args})"


@define(order=True, hash=True, frozen=True)
class MergeQuery:
    name: str
    query: Query
    only_first: bool = True

    def __str__(self) -> str:
        arr = "" if self.only_first else "[]"
        return f"{self.name}{arr}: {self.query}"

    def change_variable(self, fn: Callable[[str], str]) -> MergeQuery:
        return evolve(self, name=fn(self.name), query=self.query.change_variable(fn))


@define(order=True, hash=True, frozen=True)
class MergeTerm(Term):
    pre_filter: Term
    merge: List[MergeQuery]
    post_filter: Optional[Term] = None

    def or_merge_term(self, other: Term) -> Term:
        if isinstance(other, MergeTerm):
            return MergeTerm(
                pre_filter=self.pre_filter.or_term(other.pre_filter),
                merge=self.merge + other.merge,
                post_filter=combine_optional(self.post_filter, other.post_filter, lambda x, y: x.or_term(y)),
            )
        else:
            return evolve(self, pre_filter=self.pre_filter.or_term(other))

    def and_merge_term(self, other: Term) -> Term:
        if isinstance(other, MergeTerm):
            return MergeTerm(
                pre_filter=self.pre_filter.and_term(other.pre_filter),
                merge=self.merge + other.merge,
                post_filter=combine_optional(self.post_filter, other.post_filter, lambda x, y: x.and_term(y)),
            )
        else:
            return evolve(self, pre_filter=self.pre_filter.and_term(other))

    def __str__(self) -> str:
        merge = ", ".join(str(q) for q in self.merge)
        post = " " + str(self.post_filter) if self.post_filter else ""
        return f"{self.pre_filter} {{{merge}}}{post}"


@define(order=True, hash=True, frozen=True)
class Navigation:
    # Define the maximum level of navigation
    Max: ClassVar[int] = 250

    start: int = 1
    until: int = 1
    maybe_edge_types: Optional[List[EdgeType]] = None
    direction: str = Direction.outbound
    maybe_two_directional_outbound_edge_type: Optional[List[EdgeType]] = None

    @property
    def edge_types(self) -> List[EdgeType]:
        return self.maybe_edge_types or [EdgeTypes.default]

    def __str__(self) -> str:
        start = self.start
        until = self.until
        until_str = "" if until == Navigation.Max else until
        mo = self.maybe_two_directional_outbound_edge_type
        depth = ("" if start == 1 else f"[{start}]") if start == until and not mo else f"[{start}:{until_str}]"
        out_nav = ",".join(mo) if mo else ""
        nav = f'{",".join(self.edge_types)}{depth}{out_nav}'
        if self.direction == Direction.outbound:
            return f"-{nav}->"
        elif self.direction == Direction.inbound:
            return f"<-{nav}-"
        else:
            return f"<-{nav}->"


NavigateUntilRoot = Navigation(
    start=1, until=Navigation.Max, maybe_edge_types=[EdgeTypes.default], direction=Direction.inbound
)
NavigateUntilLeaf = Navigation(
    start=1, until=Navigation.Max, maybe_edge_types=[EdgeTypes.default], direction=Direction.outbound
)


@define(order=True, hash=True, frozen=True)
class WithClauseFilter:
    op: str
    num: int

    def __str__(self) -> str:
        if self.op == "==" and self.num == 0:
            return "empty"
        elif self.op == ">" and self.num == 0:
            return "any"
        else:
            return f"count{self.op}{self.num}"


@define(order=True, hash=True, frozen=True)
class WithClause:
    with_filter: WithClauseFilter
    navigation: Navigation
    term: Optional[Term] = None
    with_clause: Optional[WithClause] = None

    def change_variable(self, fn: Callable[[str], str]) -> WithClause:
        return evolve(
            self,
            term=self.term.change_variable(fn) if self.term else None,
            with_clause=self.with_clause.change_variable(fn) if self.with_clause else None,
        )

    def __str__(self) -> str:
        term = " " + str(self.term) if self.term else ""
        with_clause = " " + str(self.with_clause) if self.with_clause else ""
        return f"with({self.with_filter}, {self.navigation}{term}{with_clause})"


@define(order=True, hash=True, frozen=True)
class WithUsage:
    start: Union[datetime, timedelta]
    end: Union[datetime, timedelta, None]
    metrics: List[str]

    def __str__(self) -> str:
        def dd_str(d: Union[datetime, timedelta]) -> str:
            return duration_str(d) if isinstance(d, timedelta) else utc_str(d)

        end = f"::{dd_str(self.end)}" if self.end else ""
        return f'with_usage({dd_str(self.start)}{end}, [{",".join(self.metrics)}])'

    def start_from_now(self) -> datetime:
        return self.start if isinstance(self.start, datetime) else utc() - self.start

    def end_from_now(self) -> datetime:
        if self.end:
            return self.end if isinstance(self.end, datetime) else utc() - self.end
        else:
            return utc()


@define(order=True, hash=True, frozen=True)
class Limit:
    offset: int
    length: int

    def __str__(self) -> str:
        return f" limit {self.length}" if self.offset == 0 else f" limit {self.offset}, {self.length}"


# pylint: disable=not-an-iterable
@define(order=True, hash=True, frozen=True)
class Part:
    term: Term
    tag: Optional[str] = None
    with_clause: Optional[WithClause] = None
    with_usage: Optional[WithUsage] = None
    sort: List[Sort] = field(factory=list)
    limit: Optional[Limit] = None
    navigation: Optional[Navigation] = None
    reverse_result: bool = False

    def __str__(self) -> str:
        with_usage = f"{self.with_usage} " if self.with_usage is not None else ""
        with_clause = f" {self.with_clause}" if self.with_clause is not None else ""
        tag = f"#{self.tag}" if self.tag else ""
        sort = " sort " + (", ".join(f"{a.name} {a.order}" for a in self.sort)) if self.sort else ""
        limit = str(self.limit) if self.limit else ""
        nav = f" {self.navigation}" if self.navigation is not None else ""
        reverse = " reversed " if self.reverse_result else ""
        return f"{with_usage}{self.term}{with_clause}{tag}{sort}{limit}{reverse}{nav}"

    def change_variable(self, fn: Callable[[str], str]) -> Part:
        return evolve(
            self,
            term=self.term.change_variable(fn),
            with_clause=self.with_clause.change_variable(fn) if self.with_clause else None,
            sort=[sort.change_variable(fn) for sort in self.sort],
        )

    # ancestor.some_type.reported.prop -> MergeQuery
    def merge_queries_for(self, property_paths: Iterable[str]) -> List[MergeQuery]:
        def with_query_for(property_path: str) -> MergeQuery:
            try:
                assert is_ancestor_descendant(property_path)
                anc_dec, kind, _ = property_path.split(".", 2)
                direction = Direction.inbound if anc_dec == "ancestors" else Direction.outbound
                navigation = Navigation(1, Navigation.Max, direction=direction)
                subquery = Query([Part(IsTerm([kind])), Part(AllTerm(), navigation=navigation)])
                return MergeQuery(f"{anc_dec}.{kind}", subquery)
            except Exception as ex:
                raise AttributeError(
                    "The name of an ancestor variable has to follow the format: ancestors.<kind>.<path.to.variable>. "
                    "The kind defines the type of the ancestor to look for.\n"
                    "Example: ancestors.account.reported.name=test\n"
                    "Example: descendant..reported.name=test\n"
                ) from ex

        existing = {a.name: a for a in (self.term.merge if isinstance(self.term, MergeTerm) else [])}
        created = {a.name: a for a in [with_query_for(name) for name in property_paths]}
        queries = list({**created, **existing}.values())
        return queries

    def rewrite_for_ancestors_descendants(self) -> Part:
        """
        This function rewrites this part if predicates in the "magic" sections ancestors or descendants are used.
        Intention: a merge is performed by traversing the graph either inbound (ancestors) or outbound (descendants).

        Important: the merge node is found by kind only! The first matching node with correct type is merged with
        this node. The filter then is applied _after_ the node has been merged. So the filter can effectively
        filter the current node, based on properties of the merged node.

        The ancestors or descendants predicate has this form and will create a merge query:
        ancestors.<kind>.<path.to.prop> creates a merge query: {ancestors.<kind> <-[0:]- is(<kind>)}
        descendants.<kind>.<path.to.prop> creates a merge query: {descendants.<kind> -[0:]-> is(<kind>)}

        The query is rewritten in order to create a prefilter with all terms that do not depend on the merge.
        A MergeTerm is either created if not existent or the existing one will be extended with all merge query
        additions. All merge relevant parts will be performed as merge term post filter.
        Even if the query is rewritten, the logic of the query is not changed and stays the same.

        :return: the rewritten part with resolved merge parts if ancestor or descendant predicates are found.
        """

        def has_ancestor_descendant(t: Term) -> bool:
            return t.find_term(lambda trm: isinstance(trm, Predicate) and is_ancestor_descendant(trm.name)) is not None

        def ancestor_descendant_predicates(t: Term) -> List[Predicate]:
            return t.find_terms(lambda t: isinstance(t, Predicate) and is_ancestor_descendant(t.name), in_context_term=False)  # type: ignore # noqa: E501

        if has_ancestor_descendant(self.term):
            # create a filter term that is independent of the merge and execute it before the merge
            before_merge, after_merge = self.term.split_term_by(has_ancestor_descendant)
            # Create a dict here instead of a set only to ensure ordering (dict remembers order, set is not)'b
            queries = self.merge_queries_for({p.name: 1 for p in ancestor_descendant_predicates(after_merge)})
            return evolve(self, term=MergeTerm(before_merge, queries, after_merge))
        else:
            return self

    @property
    def visible_predicates(self) -> List[Predicate]:
        result: List[Predicate] = self.term.find_terms(lambda x: isinstance(x, Predicate), in_context_term=False)  # type: ignore # noqa: E501
        for ctx in self.term.find_terms(lambda x: isinstance(x, ContextTerm)):
            result.extend(ctx.visible_predicates())  # type: ignore
        return result


@define(order=True, hash=True, frozen=True)
class AggregateVariableName:
    name: str

    def __str__(self) -> str:
        return self.name

    def change_variable(self, fn: Callable[[str], str]) -> AggregateVariableName:
        return AggregateVariableName(fn(self.name))


@define(order=True, hash=True, frozen=True)
class AggregateVariableCombined:
    parts: List[Union[str, AggregateVariableName]]

    def __str__(self) -> str:
        combined = "".join(p if isinstance(p, str) else f"{{{p}}}" for p in self.parts)
        return f'"{combined}"'

    def change_variable(self, fn: Callable[[str], str]) -> AggregateVariableCombined:
        return AggregateVariableCombined(
            [p.change_variable(fn) if isinstance(p, AggregateVariableName) else p for p in self.parts]
        )


@define(order=True, hash=True, frozen=True)
class AggregateVariable:
    # name is either a simple variable name or some combination of strings and variables like "foo_{var1}_{var2}_bla"
    name: Union[AggregateVariableName, AggregateVariableCombined]
    as_name: Optional[str] = None

    def __str__(self) -> str:
        with_as = f" as {self.as_name}" if self.as_name else ""
        return f"{self.name}{with_as}"

    def all_names(self) -> List[str]:
        if isinstance(self.name, AggregateVariableCombined):
            return [avn.name for avn in self.name.parts if isinstance(avn, AggregateVariableName)]
        else:
            return [self.name.name]

    def get_as_name(self) -> str:
        def from_name() -> str:
            return self.name.name.rsplit(".", 1)[-1] if isinstance(self.name, AggregateVariableName) else str(self.name)

        return self.as_name if self.as_name else from_name()

    def change_variable(self, fn: Callable[[str], str]) -> AggregateVariable:
        return evolve(self, name=self.name.change_variable(fn))

    def property_paths(self) -> Set[str]:
        return set(self.all_names())


@define(order=True, hash=True, frozen=True)
class AggregateFunction:
    function: str
    name: Union[str, int]
    ops: List[Tuple[str, Union[int, float]]] = field(factory=list)
    as_name: Optional[str] = None

    def __str__(self) -> str:
        with_as = f" as {self.as_name}" if self.as_name else ""
        with_ops = " " + self.combined_ops() if self.ops else ""
        return f"{self.function}({self.name}{with_ops}){with_as}"

    def combined_ops(self) -> str:
        return " ".join(f"{op} {value}" for op, value in self.ops)

    def get_as_name(self) -> str:
        return self.as_name if self.as_name else f"{self.function}_of_{self.name}"

    def change_variable(self, fn: Callable[[str], str]) -> AggregateFunction:
        return evolve(self, name=fn(self.name)) if isinstance(self.name, str) else self

    def property_paths(self) -> Set[str]:
        return {self.name} if isinstance(self.name, str) else set()


@define(order=True, hash=True, frozen=True)
class Aggregate:
    group_by: List[AggregateVariable]
    group_func: List[AggregateFunction]

    def __str__(self) -> str:
        grouped = ", ".join(str(a) for a in self.group_by) + ": " if self.group_by else ""
        funcs = ", ".join(str(a) for a in self.group_func)
        return f"aggregate({grouped}{funcs})"

    def change_variable(self, fn: Callable[[str], str]) -> Aggregate:
        return Aggregate(
            [a.change_variable(fn) for a in self.group_by], [a.change_variable(fn) for a in self.group_func]
        )

    def property_paths(self) -> Set[str]:
        result = set()
        for agg in chain(self.group_by, self.group_func):
            result.update(agg.property_paths())  # type: ignore
        return result


SimpleValue = Union[str, int, float, bool]


class SortOrder:
    Asc = "asc"
    Desc = "desc"

    all = [Asc, Desc]

    @classmethod
    def reverse(cls, order: str) -> str:
        return cls.Asc if order == cls.Desc else cls.Desc


@define(order=True, hash=True, frozen=True)
class Sort:
    name: str
    order: str = SortOrder.Asc

    def __str__(self) -> str:
        return f"{self.name} {self.order}"

    def change_variable(self, fn: Callable[[str], str]) -> Sort:
        return evolve(self, name=fn(self.name))

    def reversed(self) -> Sort:
        return Sort(self.name, SortOrder.Asc if self.order == SortOrder.Desc else SortOrder.Desc)


@define(order=True, hash=True, frozen=True, slots=False)
class Query:
    parts: List[Part]
    preamble: Dict[str, SimpleValue] = field(factory=dict)
    aggregate: Optional[Aggregate] = None

    def __attrs_post_init__(self) -> None:
        if self.parts is None or len(self.parts) == 0:
            raise AttributeError(f"Expected non empty parts but got {self.parts}")

    @staticmethod
    def by(
        term: Union[str, Term], *terms: Union[str, Term], preamble: Optional[Dict[str, SimpleValue]] = None
    ) -> Query:
        res = Query.mk_term(term, *terms)
        return Query([Part(res)], preamble if preamble else {})

    def __str__(self) -> str:
        aggregate = str(self.aggregate) if self.aggregate else ""
        to_str = Predicate.value_str_rep
        preamble = "(" + ", ".join(f"{k}={to_str(v)}" for k, v in self.preamble.items()) + ")" if self.preamble else ""
        colon = ":" if self.preamble or self.aggregate else ""
        parts = " ".join(str(a) for a in reversed(self.parts))
        return f"{aggregate}{preamble}{colon}{parts}"

    @cached_property
    def merge_names(self) -> Set[str]:
        return {mt.name for part in self.parts if isinstance(part.term, MergeTerm) for mt in part.term.merge}

    @cached_property
    def merge_query_by_name(self) -> List[MergeQuery]:
        return [mt for part in self.parts if isinstance(part.term, MergeTerm) for mt in part.term.merge]

    def filter(self, term: Union[str, Term], *terms: Union[str, Term]) -> Query:
        res = Query.mk_term(term, *terms)
        parts = self.parts.copy()
        first_part = parts[0]
        if first_part.navigation is None:
            # just add the filter to this query
            parts[0] = Part(CombinedTerm(first_part.term, "and", res))
        else:
            # put to the start
            parts.insert(0, Part(res))
        return evolve(self, parts=parts)

    def filter_with(self, clause: WithClause) -> Query:
        first_part = evolve(self.parts[0], with_clause=clause)
        return evolve(self, parts=[first_part, *self.parts[1:]])

    def traverse_out(self, start: int = 1, until: int = 1, edge_type: EdgeType = EdgeTypes.default) -> Query:
        return self.traverse(start, until, edge_type, Direction.outbound)

    def traverse_in(self, start: int = 1, until: int = 1, edge_type: EdgeType = EdgeTypes.default) -> Query:
        return self.traverse(start, until, edge_type, Direction.inbound)

    def traverse_inout(self, start: int = 1, until: int = 1, edge_type: EdgeType = EdgeTypes.default) -> Query:
        return self.traverse(start, until, edge_type, Direction.any)

    def traverse(
        self, start: int, until: int, edge_type: EdgeType = EdgeTypes.default, direction: str = Direction.outbound
    ) -> Query:
        parts = self.parts.copy()
        p0 = parts[0]
        if p0.navigation:
            # we already traverse in this direction: add start and until
            if edge_type in p0.navigation.edge_types and p0.navigation.direction == direction:
                start_m = min(Navigation.Max, start + p0.navigation.start)
                until_m = min(Navigation.Max, until + p0.navigation.until)
                parts[0] = evolve(p0, navigation=evolve(p0.navigation, start=start_m, until=until_m))
            # this is another traversal: so we need to start a new part
            else:
                parts.insert(0, Part(AllTerm(), navigation=Navigation(start, until, [edge_type], direction)))
        else:
            parts[0] = evolve(p0, navigation=Navigation(start, until, [edge_type], direction))
        return evolve(self, parts=parts)

    def group_by(self, variables: List[AggregateVariable], funcs: List[AggregateFunction]) -> Query:
        aggregate = Aggregate(variables, funcs)
        return evolve(self, aggregate=aggregate)

    def set_sort(self, *sort: Sort) -> Query:
        return self.__change_current_part(lambda p: evolve(p, sort=list(sort)))

    def add_sort(self, *sort: Sort) -> Query:
        return self.__change_current_part(lambda p: evolve(p, sort=[*p.sort, *sort]))

    def with_limit(self, num: Union[Limit, int]) -> Query:
        limit = num if isinstance(num, Limit) else Limit(0, num)
        return self.__change_current_part(lambda p: evolve(p, limit=limit))

    def merge_preamble(self, preamble: Dict[str, SimpleValue]) -> Query:
        updated = {**self.preamble, **preamble} if self.preamble else preamble
        return evolve(self, preamble=updated)

    def merge_with(self, path: str, navigation: Navigation, term: Term) -> Query:
        parts = self.parts.copy()
        first_part = parts[0]
        merge = MergeQuery(path, Query([Part(term), Part(AllTerm(), navigation=navigation)]))
        term = (
            evolve(first_part.term, merge=[*first_part.term.merge, merge])
            if isinstance(first_part.term, MergeTerm)
            else MergeTerm(first_part.term, [merge])
        )
        parts[0] = evolve(first_part, term=term)
        return evolve(self, parts=parts)

    def change_variable(self, fn: Callable[[str], str]) -> Query:
        aggregate = self.aggregate.change_variable(fn) if self.aggregate else None
        parts = [p.change_variable(fn) for p in self.parts]
        return evolve(self, aggregate=aggregate, parts=parts)

    def rewrite_for_ancestors_descendants(self, additional_paths_to_select: Optional[Iterable[str]] = None) -> Query:
        def rewrite_for_additional_paths(parts: List[Part]) -> None:
            paths: Set[str] = self.aggregate.property_paths() if self.aggregate else set()
            paths.update(additional_paths_to_select or set())
            anc_desc = {path: 1 for path in paths if is_ancestor_descendant(path)}
            if anc_desc:
                current = parts[0]
                queries = current.merge_queries_for(anc_desc)
                merge_term = (
                    MergeTerm(current.term.pre_filter, queries, current.term.post_filter)
                    if isinstance(current.term, MergeTerm)
                    else MergeTerm(current.term, queries)
                )
                parts[0] = evolve(current, term=merge_term)

        adapted = [part.rewrite_for_ancestors_descendants() for part in self.parts]
        if self.aggregate or additional_paths_to_select:
            rewrite_for_additional_paths(adapted)
        return evolve(self, parts=adapted)

    def on_section(self, section: Optional[str] = PathRoot) -> Query:
        root_or_section = None if section is None or section == PathRoot else section
        absolute_section = self.change_variable(partial(variable_to_absolute, root_or_section))
        return absolute_section.rewrite_for_ancestors_descendants()

    def relative_to_section(self, section: str) -> Query:
        return self.change_variable(partial(variable_to_relative, section)) if section != PathRoot else self

    def tag(self, name: str) -> Query:
        return self.__change_current_part(lambda p: evolve(p, tag=name))

    @property
    def current_part(self) -> Part:
        # remember: the order of parts is reversed
        return self.parts[0]

    def __change_current_part(self, fn: Callable[[Part], Part]) -> Query:
        parts = self.parts.copy()
        # if navigation is defined: the current part is already defined to the end
        if parts[0].navigation:
            part = Part(AllTerm())
            parts.insert(0, part)
        else:
            part = parts[0]
        parts[0] = fn(part)
        return evolve(self, parts=parts)

    def combine(self, other: Query) -> Query:
        preamble = {**self.preamble, **other.preamble}
        if self.aggregate and other.aggregate:
            raise AttributeError("Can not combine 2 aggregations!")
        aggregate = self.aggregate if self.aggregate else other.aggregate
        left_last = self.parts[0]
        right_first = other.parts[-1]

        def combine_limit(left: Limit, right: Limit) -> Limit:
            return Limit(max(left.offset, right.offset), min(left.length, right.length))

        if left_last.navigation:
            parts = other.parts + self.parts
        else:
            if left_last.with_clause and right_first.with_clause:
                raise AttributeError("Can not combine 2 with clauses!")
            term = left_last.term & right_first.term
            if left_last.tag and right_first.tag:
                raise AttributeError("Can not combine 2 tag clauses!")
            tag = left_last.tag if left_last.tag else right_first.tag
            with_clause = left_last.with_clause if left_last.with_clause else right_first.with_clause
            with_usage = left_last.with_usage if left_last.with_usage else right_first.with_usage
            sort = combine_optional(left_last.sort, right_first.sort, lambda m, r: m + r)
            limit = combine_optional(left_last.limit, right_first.limit, combine_limit)
            combined = Part(term, tag, with_clause, with_usage, sort if sort else [], limit, right_first.navigation)
            parts = [*other.parts[0:-1], combined, *self.parts[1:]]
        return Query(parts, preamble, aggregate)

    @property
    def visible_predicates(self) -> List[Predicate]:
        """
        Returns a list of all predicates in this query.
        """
        return [pred for part in self.parts for pred in part.visible_predicates]

    def find_terms(self, fn: Callable[[Term], bool], **kwargs: bool) -> List[Term]:
        return [t for p in self.parts for t in p.term.find_terms(fn, **kwargs)]

    def analytics(self) -> Tuple[Dict[str, int], Dict[str, List[str]]]:
        counters: Dict[str, int] = defaultdict(lambda: 0)
        names: Dict[str, List[str]] = defaultdict(list)

        def term_analytics(term: Term) -> None:
            name = type(term).__name__
            counters[f"term_{name.lower()}"] += 1
            if isinstance(term, Predicate):
                counters[f"op_{term.op}"] += 1
                names["predicate_names"].append(term.name)
            elif isinstance(term, CombinedTerm):
                term_analytics(term.left)
                term_analytics(term.right)

        def with_clause_analytics(clause: WithClause) -> None:
            counters["with_clause"] += 1
            if clause.term:
                term_analytics(clause.term)
            if clause.navigation:
                navigation_analytics(clause.navigation)
            if clause.with_clause:
                with_clause_analytics(clause.with_clause)

        def navigation_analytics(navigation: Navigation) -> None:
            counters["navigation"] += 1
            counters[f"navigation_{navigation.direction}"] += 1
            counters[f"navigation_{navigation.edge_types}"] += 1
            counters["navigation_until_max"] = max(counters["navigation_until_max"], navigation.until)

        def is_ancestor_merge(q: Query) -> bool:
            return (
                len(q.parts) == 2
                and q.aggregate is None
                and q.parts[1].navigation is not None
                and q.parts[1].navigation.direction == "in"
                and q.parts[1].navigation.until > 1
                and isinstance(q.parts[0].term, IsTerm)
            )

        def query_analytics(q: Query) -> None:
            if q.preamble:
                names["preamble_keys"].extend(q.preamble.keys())
            if q.aggregate:
                if q.aggregate.group_by:
                    names["aggregate_by"].extend(str(gb.name) for gb in q.aggregate.group_by)
                    counters["aggregate_by"] += len(q.aggregate.group_by)
                if q.aggregate.group_func:
                    names["aggregate_func"].extend(str(gb.name) for gb in q.aggregate.group_func)
                    counters["aggregate_func"] += len(q.aggregate.group_func)
            for part in q.parts:
                if isinstance(part.term, MergeTerm):
                    term_analytics(part.term.pre_filter)
                    counters["merge_terms"] += 1
                    for merge in part.term.merge:
                        names["merge_names"].append(merge.name)
                        counter_name = "merge_ancestors_by_kind" if is_ancestor_merge(merge.query) else "merge_other"
                        query_analytics(merge.query)
                        counters[counter_name] += 1
                    if part.term.post_filter:
                        term_analytics(part.term.post_filter)
                else:
                    term_analytics(part.term)
                if part.limit:
                    counters["limits"] += 1
                if part.sort:
                    counters["sorts"] += 1
                    names["sort_names"].extend(sort.name for sort in part.sort)
                if part.navigation:
                    navigation_analytics(part.navigation)
                if part.with_clause:
                    with_clause_analytics(part.with_clause)

        query_analytics(self)

        return counters, names

    @staticmethod
    def mk_term(term: Union[str, Term], *args: Union[str, Term]) -> Term:
        def make_term(t: Union[str, Term]) -> Term:
            if isinstance(t, Term):
                return t
            elif isinstance(t, str):
                return IsTerm([t])
            else:
                raise AttributeError(f"Expected term or string, but got {t}")

        term_in = list(args)
        term_in.insert(0, term)
        terms = map(make_term, term_in)
        # noinspection PyTypeChecker
        return reduce(lambda left, right: CombinedTerm(left, "and", right), terms)

    def structure(self) -> Json:
        """
        External representation of the query.
        """

        def merge_query_structure(m: MergeQuery) -> Json:
            return {"name": m.name, "query": query_structure(m.query), "only_first": m.only_first}

        def term_structure(term: Term) -> Json:
            if isinstance(term, AllTerm):
                return {"kind": "all"}
            elif isinstance(term, CombinedTerm):
                return {
                    "kind": "combined",
                    "left": term_structure(term.left),
                    "op": term.op,
                    "right": term_structure(term.right),
                }
            elif isinstance(term, ContextTerm):
                return {"kind": "context", "name": term.name, "term": term_structure(term.term)}
            elif isinstance(term, FulltextTerm):
                return {"kind": "fulltext", "text": term.text}
            elif isinstance(term, FunctionTerm):
                return {"kind": "function", "property_path": term.property_path, "args": term.args}
            elif isinstance(term, IdTerm):
                return {"kind": "id", "id": term.id}
            elif isinstance(term, IsTerm):
                return {"kind": "is", "kinds": term.kinds}
            elif isinstance(term, MergeTerm):
                result = {
                    "kind": "merge",
                    "pre_filter": term_structure(term.pre_filter),
                    "merge": [merge_query_structure(m) for m in term.merge],
                }
                if term.post_filter:
                    result["post_filter"] = term_structure(term.post_filter)
                return result
            elif isinstance(term, NotTerm):
                return {"kind": "not", "term": term_structure(term.term)}
            elif isinstance(term, Predicate):
                return {"kind": "predicate", "name": term.name, "op": term.op, "value": term.value, "args": term.args}
            else:
                raise AttributeError(f"Unknown term kind {term}")

        def sort_structure(sort: Sort) -> Json:
            return {"name": sort.name, "order": sort.order}

        def limit_structure(limit: Limit) -> Json:
            return {"offset": limit.offset, "length": limit.length}

        def navigation_structure(navigation: Navigation) -> Json:
            return {
                "direction": navigation.direction,
                "edge_types": navigation.edge_types,
                "start": navigation.start,
                "until": navigation.until,
            }

        def with_clause_structure(clause: WithClause) -> Json:
            return {
                "term": term_structure(clause.term) if clause.term else None,
                "navigation": navigation_structure(clause.navigation) if clause.navigation else None,
                "with": with_clause_structure(clause.with_clause) if clause.with_clause else None,
            }

        def part_structure(part: Part) -> Json:
            return {
                "term": term_structure(part.term),
                "tag": part.tag,
                "with": with_clause_structure(part.with_clause) if part.with_clause else None,
                "sort": [sort_structure(sort) for sort in part.sort],
                "limit": limit_structure(part.limit) if part.limit else None,
                "navigation": navigation_structure(part.navigation) if part.navigation else None,
            }

        def aggregate_variable_structure(av: AggregateVariable) -> Json:
            if isinstance(av.name, AggregateVariableName):
                return {"name": av.name.name, "as": av.get_as_name()}
            elif isinstance(av.name, AggregateVariableCombined):
                return {
                    "combined_names": [p if isinstance(p, str) else {"name": p.name} for p in av.name.parts],
                    "as": av.get_as_name(),
                }
            else:
                raise AttributeError(f"Unknown aggregate variable name type {av.name}")

        def aggregate_function_structure(av: AggregateFunction) -> Json:
            return {"function": av.function, "name": av.name, "ops": av.ops, "as": av.as_name}

        def aggregate_structure(aggregate: Aggregate) -> Json:
            return {
                "group_by": [aggregate_variable_structure(gb) for gb in aggregate.group_by],
                "group_func": [aggregate_function_structure(gf) for gf in aggregate.group_func],
            }

        def query_structure(q: Query) -> Json:
            return {
                "preamble": q.preamble,
                "parts": [part_structure(part) for part in reversed(q.parts)],
                "aggregate": aggregate_structure(q.aggregate) if q.aggregate else None,
            }

        return query_structure(self)


# register serializer for this class
set_deserializer(Term.from_json, Term)
